/*
**  makerj7.c
**
**  mg - July 1995
*/

#define G_EXTERN extern

#include "defs.h"

void add_to_bigr_set(int);
void remove_from_bigr_set(int);
void add_to_seen_set(int);
void update_user_rels(char *, char *, int, int, int, int, int);

G_EXTERN void write_termset_file(struct query_terms *, int);
G_EXTERN void find_docset(char *, int *, int *);
G_EXTERN void extract_terms(int, int, int, int);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void write_rels_file(int);

void make_rj(char docid[], char *doc_date,	int set_recno,
						 int abs_recno, int rj, int offset, int length,
						 int full_doclen, int part_doclen, int previous_rel) {
  /*
  **  Make relevance judgement on a document.
	**  The termset will be amended according to the values
	**  passed in (previous_rel, rj)
	**
  **     P = previous_rel,  C = current rel (rj)
	**
	**     P IN [0, 1, 2, 3]       0  no previous judgement
	**                             1  Full doc
	**                             2  Passage only
	**                             3  Not relevant
	**
	**  P   C   Meaning   Action                      BIGR       SEEN
	**  -   -   -------   ------------------------    ----       ----
	
	a)  0   F   0 -> F    Merge full doc terms         ADD        ADD
	                      with current termset
	b)  0   P   0 -> P    Merge passage terms          ADD        ADD
	                      with current termset
	c)  0   N   0 -> N    Do nothing                   ---        ADD
	  
	d)  1   F   F -> F    Do nothing                   ---        ---
	e)  1   P   F -> P    a) little_r-- for terms      ---        ---
	                         in full doc
	                      b) little_r++ for terms
	                         in passage
	f)  1   N   F -> N    a) little_r-- for terms      REM        ---
	                         in full doc
	  
	g)  2   F   P -> F    Merge not-passage terms      ---        ---
	                      with current termset
	h)  2   P   P -> P    Do nothing                   ---        ---
	i)  2   N   P -> N    a) little_r-- for terms      REM        ---
	                         in passage
	  
	j)  3   F   N -> F    Same as 0 -> F               ADD        ---
	k)  3   P   N -> P    Same as 0 -> P               ADD        ---
	l)  3   N   N -> N    Same as 0 -> N               ---        ---

	**
  */

	int debugging = atoi(getenv("MAKERJ_DEBUG"));

  int doc_set;
  int np;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];

  /*
  **  First write the existing termset to file so that it is
  **  correct for the extract_terms() function. This function
  **  sorts the terms into ascending parsed order before
  **  writing them to disc.
  */

  write_termset_file(termset, TERM_NO);

	if (debugging) {
		switch (previous_rel) {
      case 0 :
				fprintf(stderr, "make_rj() - previous_rel(0) - no previous rel\n");
				break;
      case 1 :
				fprintf(stderr, "make_rj() - previous_rel(1) - full record\n");
				break;
      case 2 :
				fprintf(stderr, "make_rj() - previous_rel(2) - passage only\n");
				break;
      case 3 :
				fprintf(stderr, "make_rj() - previous_rel(3) - not relevant\n");
				break;
		}
	}

  
  /*
  **  The appropriate terms are extracted and written to "fulldoc_file",
	**  "passage_file" and "full_minus_pass_file" in "show_record()". The
	**  last file contains terms in the full document but not in the 
	**  passage for use when converting F->P or P->F
	**
  **  First find set for document to enable adjustment of
	**  BIGR_SET and SEEN_SET where appropriate.
  */

	find_docset(docid, &doc_set, &np);

	if (debugging) {
		sprintf(bss_command, "show f=0 s=%d r=0", doc_set);

		if ((return_code = i0(bss_command, bss_result)) != 0) {
			terminate_run("make_rj()", bss_command, return_code);
		}

		fprintf(stderr, "make_rj():: docset = [%d], doc = [%s]\n",
						doc_set, bss_result);
	}


  /*
  **  Update BIGR_SET & SEEN_SET as appropriate
  */

	switch (previous_rel) {

    case 0 : case 3 :

			if (rj != R_NOT) {
				add_to_bigr_set(doc_set);
			}

			if (previous_rel == 0) {
				add_to_seen_set(doc_set);
			}

			break;

  	case 1 : case 2 :

			if (rj == R_NOT) {
				remove_from_bigr_set(doc_set);
			}

			break;
	}

  /*
  **  Don't need doc_set anymore so delete it.
  */
  
  sprintf(bss_command, "delete %d", doc_set);
  
	if ((return_code = i0(bss_command, bss_result)) != 0) {
    terminate_run("make_rj()", bss_command, return_code);
  }

	/*
  **  The majority of options involve F or P judgements so set
	**  set NEW_ITERATION by default to TRUE.
	**
	**  It will be set to FALSE where nothing is done,
	**  i.e. 0->N, N->N, F->F and P->P
	*/

	NEW_ITERATION = TRUE;


  switch (rj) {

    case R_FULL :

			switch (previous_rel) {

			  case 0 : case 3 :
					/*
					**  0 -> F, N -> F
					**
					**  Merge "fulldoc_file" terms with current termset.
					*/
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

			  case 1 :
					/*
					**  F -> F
					**
					**  Do nothing
					*/
					NEW_ITERATION = FALSE;
					break;

			  case 2 :
					/*
					**  P -> F
					**
					**  Merge "full_minus_pass_file" with current termset.
					*/
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

			}

			break;

    case R_PASSAGE :

			switch (previous_rel) {

			  case 0 : case 3 :
					/*
					**  0 -> P, N -> P
					**  Merge "passage_file" terms with current termset
					*/
				
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

			  case 1 :
					/*
					**  F -> P
					**  Need to decrement little_r for terms in
					**  "full_minus_pass_file"
					**  i.e. do an R_NOT with this file.
					*/
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

			  case 2 :
					/*
					**  P -> P
					**  Do nothing
					*/
					NEW_ITERATION = FALSE;
					break;

			}

			break;

	  case R_NOT :

			switch (previous_rel) {

			  case 0 : case 3 :
					/*
					**  0 -> N or N -> N
					**  Do nothing
					*/
					NEW_ITERATION = FALSE;
					break;

			  case 1 :
					/*
					**  F -> N
					**  Decrement little_r for all terms in "fulldoc_file"
					**  This is done in extract_terms (in extract.c)
					*/
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

			  case 2 :
					/*
					**  P -> N
					**  Decrement little_r for all terms in "passage_file" (i.e.
					**  those terms only in the passage).
					**  This is done in extract_terms (in extract.c)
					*/
					extract_terms(rj, previous_rel, full_doclen, part_doclen);
					break;

      }

			break;
  }
	
	/*
  **  Now update the user_rels structure.
	*/

	update_user_rels(docid, doc_date, set_recno, abs_recno, rj, offset, length);
}


void add_to_bigr_set(int doc_set) {
	int debugging = atoi(getenv("ADD_TO_BIGR_SET_DEBUG"));

	int new_bigR_set = NONE_ASSIGNED;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
	char *rbuf;

	if (debugging) {
		fprintf(stderr,
						"add_to_bigr_set(1) :: BIGR_SET = %d, BIGR = %d\n",
						BIGR_SET, BIGR);
	}

	sprintf(bss_command, "find s=%d s=%d op=or", BIGR_SET, doc_set);
	
	if ((return_code = i0(bss_command, bss_result)) != 0) {
		terminate_run("add_to_bigr_set()", bss_command, return_code);
	}
	
	rbuf = bss_result;
	sscanf(rbuf, "S%d np=%d", &new_bigR_set, &BIGR);

  if (new_bigR_set != NONE_ASSIGNED) {
    /*
    **  delete old BIGR_SET then set BIGR_SET to new_bigR_set
    */

    sprintf(bss_command, "delete %d", BIGR_SET);

    if ((return_code = i0(bss_command, bss_result)) != 0) {
      terminate_run("add_to_bigr_set()", bss_command, return_code);
    }
		
    BIGR_SET = new_bigR_set;

		if (debugging) {
			fprintf(stderr,
							"add_to_bigr_set(2) :: BIGR_SET = %d, BIGR = %d\n",
							BIGR_SET, BIGR);
		}
	}
	else {
		fprintf(stderr, "Problem in adding to BIGR_SET\n");
	}
}


void remove_from_bigr_set(int doc_set) {
	int debugging = atoi(getenv("REMOVE_FROM_BIGR_SET_DEBUG"));

	int new_bigR_set = NONE_ASSIGNED;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
	char *rbuf;

	if (debugging) {
		fprintf(stderr,
						"remove_from_bigr_set(1) :: BIGR_SET = %d, BIGR = %d\n",
						BIGR_SET, BIGR);
	}

	sprintf(bss_command, "find s=%d s=%d op=not", BIGR_SET, doc_set);
				
	if ((return_code = i0(bss_command, bss_result)) != 0) {
		terminate_run("remove_from_bigr_set()", bss_command, return_code);
	}
				
	rbuf = bss_result;
	sscanf(rbuf, "S%d np=%d", &new_bigR_set, &BIGR);

  if (new_bigR_set != NONE_ASSIGNED) {
    /*
    **  delete old BIGR_SET then set BIGR_SET to new_bigR_set
    */

    sprintf(bss_command, "delete %d", BIGR_SET);

    if ((return_code = i0(bss_command, bss_result)) != 0) {
      terminate_run("remove_from_bigr_set()", bss_command, return_code);
    }
		
    BIGR_SET = new_bigR_set;

		if (debugging) {
			fprintf(stderr,
							"remove_from_bigr_set(2) :: BIGR_SET = %d, BIGR = %d\n",
							BIGR_SET, BIGR);
		}
	}
	else {
		fprintf(stderr, "Problem removing doc from BIGR_SET\n");
	}
}


void add_to_seen_set(int doc_set)
{
	int debugging = atoi(getenv("ADD_TO_SEEN_SET_DEBUG"));

	int new_seen_set = NONE_ASSIGNED;
	int np;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
	char *rbuf;

	if (debugging) {
		fprintf(stderr,
						"add_to_seen_set(1) :: SEEN_SET = %d, NO_SEEN = %d\n",
						SEEN_SET, NO_SEEN);
	}

  /*
  **  Now check to see if already in SEEN_SET. If not add to user_rels[]
  **
  **  Test np. If np == 0 then it wasn't in the BIGR_SET but it might
  **  be in SEEN_SET since it could have been judged as NO
  */
  
  sprintf(bss_command, "find s=%d s=%d op=and save=n", doc_set, SEEN_SET);

  if ((return_code = i0(bss_command, bss_result)) != 0) {
    terminate_run("add_to_seen_set()", bss_command, return_code);
  }

  rbuf = bss_result;
  sscanf(rbuf, "np=%d", &np);

  if (np == 0) {
    /*
    **  Not there
    **
    **  Now add document to SEEN_SET
    */

    sprintf(bss_command, "f s=%d s=%d op=or", SEEN_SET, doc_set);
    
    if ((return_code = i0(bss_command, bss_result)) != 0) {
      terminate_run("add_to_seen_set()", bss_command, return_code);
    }
    
    rbuf = bss_result;
    sscanf(rbuf, "S%d np=%d", &new_seen_set, &NO_SEEN);
    
    /*
    **  delete old 'seen_set'
    */
    
    sprintf(bss_command, "delete %d", SEEN_SET);
    
    if ((return_code = i0(bss_command, bss_result)) != 0) {
      terminate_run("add_to_seen_set()", bss_command, return_code);
    }
    
    /*
    **  set SEEN_SET = new_seen_set
    */

    SEEN_SET = new_seen_set;
  }

	if (debugging) {
		fprintf(stderr,
						"add_to_seen_set(1) :: SEEN_SET = %d, NO_SEEN = %d\n",
						SEEN_SET, NO_SEEN);
	}
}


void update_user_rels(char *docid, char *doc_date,
											int set_recno, int abs_recno, int rj,
											int offset, int length)
{
  /*
  **
  **  Record in user_rels[] if not already seen.
	**  Otherwise update (set_recno, weight, rj, offset, length) values
	**
  **  First set all old user_rels sets to the new SEEN_SET
  **  While doing this check to see if it's already recorded.
  **  If so change update its rj field.
  */

	int debugging = atoi(getenv("UPDATE_USER_RELS_DEBUG"));
	int last_record = atoi(getenv("MAX_RECS_TO_SHOW"));

  int next_rel;
	int there = FALSE;

	if (USER_RELS_COUNT > 0) {
		for (next_rel = 0; next_rel < USER_RELS_COUNT; next_rel++) {
			(user_rels + next_rel)->set = SEEN_SET;
			
			if (there = ((user_rels + next_rel)->abs_recno == abs_recno)) {
				/*
				**  Already recorded details about this document. Update
				**  current values of (set_recno, weight, rj, offset, length)
				*/
				
				(user_rels + next_rel)->set_recno = set_recno;
				(user_rels + next_rel)->weight = SHOW_WGT;
				(user_rels + next_rel)->rj = rj;
				(user_rels + next_rel)->offset = offset;
				(user_rels + next_rel)->length = length;
				break;
			}
		}
	}
  
  /*
  **  If it is not already in user_rels[] add it.
  */
  
	if (!there) {
    (user_rels + USER_RELS_COUNT)->relno = RELJ_COUNT++;
    sprintf((user_rels + USER_RELS_COUNT)->docid, "%s", docid);
    (user_rels + USER_RELS_COUNT)->set = SEEN_SET;
    (user_rels + USER_RELS_COUNT)->abs_recno = abs_recno;
    (user_rels + USER_RELS_COUNT)->set_recno = set_recno;
    (user_rels + USER_RELS_COUNT)->weight = SHOW_WGT;
    (user_rels + USER_RELS_COUNT)->rj = rj;
    (user_rels + USER_RELS_COUNT)->offset = offset;
    (user_rels + USER_RELS_COUNT)->length = length;
    (user_rels + USER_RELS_COUNT)->sflag = 'C';
		sprintf((user_rels + USER_RELS_COUNT)->date, doc_date);

    /*
    **  Now search for title in hit_titles[]
    */

		for (next_rel = 0; next_rel < last_record; next_rel++) {
			if (hit_titles[next_rel].abs_recno ==
					(user_rels + USER_RELS_COUNT)->abs_recno) {
				sprintf((user_rels + USER_RELS_COUNT)->title,
								"%s", hit_titles[next_rel].title);
				break;
			}
		}
		
    USER_RELS_COUNT++;
  }

  /*
  **  Write to user_rels file - this function first
  **  sorts the rels into relno order 
  */
	
  write_rels_file(USER_RELS_COUNT);
}
